import json
import thread
import logging

import requests
import websocket

import horse.config


class SocketServer(websocket.WebSocketApp):

    def __init__(self, jockey):
        response = requests.get('https://slack.com/api/rtm.start', params={
            "token": horse.config.SLACK_API_TOKEN
        })
        self.metadata = response.json()
        self.url = self.metadata['url']
        self.keep_running = True
        self.last_ping_tm = 0
        self.header = []
        self.cookie = None
        self.subprotocols = None
        self.sock = None
        self.get_mask_key = None
        self.jockey = jockey

    def start(self):
        def run(*args):
            self.run_forever()

        logging.info('Starting socket client on new thread.')
        thread.start_new_thread(run, ())

    def on_open(self, socket):
        logging.info('Socket client open')

    def on_close(self, socket):
        logging.info('Socket client closed. Reopening.')
        self.start()

    def on_message(self, socket, message):
        data = json.loads(message)
        self.jockey.handle_socket(data)

    def on_ping(self, socket, ping):
        logging.info('Socket PING')

    def on_pong(self, socket, pong):
        logging.info('Socket PONG')

    def on_cont_message(self, socket):
        logging.info('Socket CONT')

    def on_error(self, socket, error):
        logging.error('Socket ERROR: {0}'.format(error))



from horse.bridles.base import CommandBridle


class Inventory(CommandBridle):

    class Meta(CommandBridle.Meta):
        command = 'inventory'
        display_name = 'Horse Help'
        display_icon = 'http://ryangrieve.com/labs/slack_icons/help.png'
        description = "Lists all bridles currently available."
        help_text = ["Usage: `/horse inventory`"]

    def execute(self, user, channel, operands):
        bridles = ", ".join(
            '_%s_' % x.Meta.command
            for x in self.jockey.bridles['command']
            if not x.Meta.secret
            and x.Meta.command != ""
            and not x.Meta.command.startswith('_')
        )
        self.message(
            channel,
            "Horse has loaded the following bridles: %s" % bridles
        )


class Help(CommandBridle):

    class Meta(CommandBridle.Meta):
        command = 'help'
        display_name = 'Horse Help'
        display_icon = 'http://ryangrieve.com/labs/slack_icons/help.png'
        description = "Prints help and usage information on available bridles"
        help_text = ["Usage: `/horse help [bridle-name]`"]

    def execute(self, user, channel, operands):
        if len(operands) > 0:
            bridle_name = operands[0]
            bridle = None
            for m in self.jockey.bridles['command']:
                if m.__class__.__name__ == bridle_name:
                    bridle = m
                elif m.Meta.command == bridle_name:
                    bridle = m
            if bridle:
                msg = "%s - %s" % (
                    bridle.__class__.__name__,
                    bridle.Meta.description
                )
                for line in bridle.Meta.help_text:
                    msg += "\n\t- %s" % line
                self.message(channel, msg)
            else:
                self.message(channel, "Bridle '%s' not found" % bridle_name)
        else:
            self.message(channel, self.Meta.help_text)
            self.message(
                channel,
                "(Hint - Try typing `/horse inventory` for a list of bridles)"
            )



import re
import six


class BridleBase(type):

    class DefaultMeta(object):
        pass

    def __new__(cls, name, bases, attrs):
        meta = attrs.pop('Meta', None)
        new_class = super(BridleBase, cls).__new__(cls, name, bases, attrs)
        if meta is None:
            meta = BridleBase.DefaultMeta
        if 'abstract' not in meta.__dict__:
            meta.abstract = False
        new_class.Meta = meta
        return new_class


class Bridle(six.with_metaclass(BridleBase)):

    class Meta:
        abstract = True
        display_name = 'Horse'
        display_icon = 'http://ryangrieve.com/labs/slack_icons/horse.png'
        description = 'No description provided'

    def __init__(self, jockey):
        self.jockey = jockey

    def message(self, channel, message, **kwargs):
        if 'username' not in kwargs:
            kwargs['username'] = self.Meta.display_name
        if 'icon_url' not in kwargs:
            kwargs['icon_url'] = self.Meta.display_icon

        self.jockey.slack.send_message(
            channel=channel,
            text=message,
            **kwargs
        )


class CommandBridle(Bridle):

    class Meta(Bridle.Meta):
        abstract = True
        command = ''
        help_text = ['No help text provided']
        secret = False

    def execute(self, user, channel, operands):
        return ""


class ListenerBridle(Bridle):

    class Meta(Bridle.Meta):
        abstract = True
        regex = ''

    def __init__(self, jockey):
        super(ListenerBridle, self).__init__(jockey)
        self._pattern = re.compile(self.Meta.regex)

    def execute(self, user, channel, context):
        return ""


class WebhookBridle(Bridle):

    class Meta(Bridle.Meta):
        abstract = True
        method = 'GET'
        path = ''

    def execute(self, path, data):
        return ""


class EventBridle(Bridle):

    class Meta(Bridle.Meta):
        abstract = True
        event = ''

    def execute(self, data):
        return ""



import os
import importlib

import horse.config
import horse.jockey
import horse.ext  # NOQA


def run(**kwargs):
    config_path = os.getenv('HORSE_CONFIG_MODULE', None)
    if config_path is None:
        raise Exception(
            'No configuration set. Use enviroment variable HORSE_CONFIG_MODULE'
        )
    config_module = importlib.import_module(config_path)
    horse.config.__dict__.update(config_module.__dict__)
    jockey = horse.jockey.Jockey()
    jockey.start()



import logging

import requests


class API(object):

    api_root = 'https://slack.com/api/{0}'

    def __init__(self, api_token):
        logging.info('Slack API client started with {0}'.format(api_token))
        self._token = api_token

    def _make_request(self, endpoint, **params):
        params['token'] = self._token
        response = requests.post(self.api_root.format(endpoint), data=params)
        logging.info(
            'Request: {0} - {1}'.format(endpoint, response.status_code)
        )
        if response.status_code != 200:
            logging.error(response.content)
        return response

    def send_message(self, **kwargs):
        return self._make_request('chat.postMessage', **kwargs)



from ..base import EventBridle


# Inherit from EventBridle class to have your code executed when slack sends
# a specific event type via the realtime api. Event types are listed here:
#     https://api.slack.com/rtm

# To deal with 'message' type events, use the ListenerBridle instead.
class SimpleEventBridle(EventBridle):

    # Inherit from the parents Meta and add the event type you wish to act on
    class Meta(EventBridle.Meta):
        event = 'hello'

    # The execute function is called when the specific event type arrives.
    # The entirety of the event message is passed as a parameters to this
    # function. For more details on the data contents, see the Slack docs
    # linked above.
    def execute(self, data):
        self.message('#horsebox', "Horse is connected via web sockets!")



import sys


class HorseExtensions(object):

    def __getattr__(self, value):
        import importlib
        try:
            return importlib.import_module('horse_' + value)
        except ImportError:
            raise ImportError(
                "Cannot find horse extension 'horse.ext.{0}'".format(
                    value
                )
            )


sys.modules[__name__] = HorseExtensions()



from ..base import CommandBridle


# Inherit from CommandBridle class to implement a slash command bridle
# this command is executed using `/horse simple_test`
class SimpleCommand(CommandBridle):

    # Inherit from the parents Meta and add command word
    class Meta(CommandBridle.Meta):
        command = 'simple_test'

    # The execute function is called by the jockey when the command word
    # is used. The user and channel are passed to the function along with
    # any additional operands that were provided to the command.

    # For example, if this command was invoked like:
    #     `/horse simple_test 123 apples peach`
    # then operands would be:
    #     ["123", "apples", "peach"].

    # All operands are strings.
    def execute(self, user, channel, operands):
        # You can pass either the user or channel id to the `message` method
        # to reply publicly or in a PM
        self.message(channel, "Message visible in channel")
        self.message(user, "Message is PM'd to user")

        # Any string value returned by the method will be shown as a "gray"
        # message in the channel, and only visible to the user who invoked
        # the command.
        return 'Message in channel, but only visible to user'



import logging

import sqlobject
from sqlobject.dberrors import OperationalError

import horse.config
import horse.models


def configure():
    db = sqlobject.connectionForURI(horse.config.DATABASE_URI)
    sqlobject.sqlhub.processConnection = db
    for model in [horse.models.User, horse.models.Channel]:
        try:
            model.createTable()
        except OperationalError:
            pass
        else:
            logging.info('\t{0} table created'.format(model.__name__))


def verify_models(bridle_models):
    logging.info('Verifying DB tables for loaded models')
    for model in bridle_models:
        try:
            model.createTable()
        except OperationalError:
            pass
        else:
            logging.info('\t{0} table created'.format(model.__name__))



from ..base import WebhookBridle


# Inherit from WebhookBridle if you want your code executed when a specific
# URL on the horse instance is requested. This is useful for tying in
# functionality from other services, like JIRA, Git or CI.
class SimpleWebhookBridle(WebhookBridle):

    # Inherit from parents Meta and provide path, and optionaly method. Method
    # can be 'GET' or 'POST', and defaults to 'GET' if it's missing.
    class Meta(WebhookBridle.Meta):
        path = 'testing'
        method = 'GET'

    # The execute function is called whenever the specified path is requested
    # via the specified method. It is passed the path and the parameters it
    # was requested with. Both form encoded and JSON parameters are supported
    # and automatically decoded.
    def execute(self, path, params):
        self.message('#horsebox', "webhooks!")



import datetime

from horse.bridles.base import CommandBridle


class Uptime(CommandBridle):

    class Meta(CommandBridle.Meta):
        command = 'uptime'

    def execute(self, user, channel, operands):
        response = '> Uptime {0}'.format(
            datetime.datetime.now() - self.jockey.start_time
        )
        self.message(channel, response)



import sqlobject

from ..base import CommandBridle
from ..base import BridleModel


# Any classes that inherit from BridleModel will automatically handled by
# jockey and equivalent tables made in the managed database. These classes
# are derived from sqlobject.SQLObject.
class SimpleCounter(BridleModel):
    user = sqlobject.StringCol()
    channel = sqlobject.StringCol()


class PersistCommand(CommandBridle):

    class Meta(CommandBridle.Meta):
        command = 'persist'

    def execute(self, user, channel, operands):
        # create an instance of the class to save a record to the database
        SimpleCounter(user=user, channel=channel)

        # query the database for num records, see sqlobject docs for usage
        num_records = SimpleCounter.select().count()
        self.message(channel, "This has been called {0} time(s)".format(
            num_records
        ))



import json
import thread
import logging

from flask import Flask
from flask import request


class HTTPServer(object):

    FORM_DATA_MIME = 'application/x-www-form-urlencoded'
    JSON_DATA_MIME = 'application/json'

    def __init__(self, jockey):
        self.jockey = jockey
        self.app = Flask('horse-http-server')
        self.app.add_url_rule(
            '/',
            'handler',
            self.handle_request,
            methods=['GET', 'POST'],
            defaults={'path': ''}
        )
        self.app.add_url_rule(
            '/<path:path>',
            'handler',
            self.handle_request,
            methods=['GET', 'POST']
        )

    def handle_request(self, path):
        logging.info('Inbound request: {0}'.format(path))
        if path == 'command':
            response = self.jockey.handle_command(request.form)
        else:
            if request.method == 'GET':
                response = self.jockey.handle_http_get(path, request.args)
            elif request.method == 'POST':
                if request.headers['Content-Type'] == self.FORM_DATA_MIME:
                    response = self.jockey.handle_http_post(path, request.form)
                elif request.headers['Content-Type'] == self.JSON_DATA_MIME:
                    data = json.loads(request.body)
                    response = self.jockey.handle_http_post(path, data)
        logging.info('Outbound response: {0}'.format(response))
        if response is None:
            response = ""
        return str(response)

    def start(self):
        def run(*args):
            self.app.run(host="0.0.0.0", port=5000)

        logging.info('Starting HTTP server on new thread.')
        thread.start_new_thread(run, ())



import json
import requests

from horse import config


def shorten(longurl):
    endpoint = "https://www.googleapis.com/urlshortener/v1/url?key={0}".format(
        config.GAPI_API_TOKEN
    )

    response = requests.post(
        url=endpoint,
        headers={
            "Content-Type": "application/json"
        },
        data=json.dumps({
            "longUrl": longurl
        })
    )

    return response.json()['id']



from ..base import ListenerBridle


# Inherit from ListenerBridle to listen for specific text patterns in user
# messages. This will only work for public channels, or groups that Horse (or
# rather, the api key owner) is a member of.
class SimpleListenerBridle(ListenerBridle):

    # Inherit from parents Meta and add regex pattern to match
    class Meta(ListenerBridle.Meta):
        regex = 'test(ing)?'

    # The execute function is called when a message matches the provided regex
    # pattern. It is passed the user, the channel, the re.MatchObject of the
    # match, and the context (the full text of the message).
    def execute(self, user, channel, match, context):
        self.message(channel, "You said: {0}".format(match.group()))



import types
import inspect
import logging
import datetime
import traceback
import importlib


import horse.config
import horse.slack
import horse.servers.http
import horse.servers.socket
import horse.database
import horse.models
import horse.bridles.base


class Jockey(object):

    def __init__(self):
        logging.info('Jockey initialising.')
        self.socket = horse.servers.socket.SocketServer(self)
        self.http = horse.servers.http.HTTPServer(self)
        self.slack = horse.slack.API(horse.config.SLACK_API_TOKEN)
        horse.database.configure()
        self.load_metadata()
        self.load_bridles()
        logging.info('Jockey ready.')

    def start(self):
        logging.info('Jockey starting.')
        self.socket.start()
        self.http.start()
        self.start_time = datetime.datetime.now()
        logging.info('Jockey started.')
        while True:
            pass

    def get_user(self, id):
        query = horse.models.User.selectBy(string_id=id)
        results = list(query)
        if len(results) > 0:
            return results[0]
        else:
            return None

    def get_channel(self, id):
        query = horse.models.Channel.selectBy(string_id=id)
        results = list(query)
        if len(results) > 0:
            return results[0]
        else:
            return None

    def load_metadata(self):
        for user in self.socket.metadata['users']:
            if user['deleted']:
                continue
            if not self.get_user(user['id']):
                horse.models.User(
                    string_id=user['id'],
                    username=user['name'],
                    real_name=user['profile']['real_name_normalized'],
                    image=user['profile']['image_72']
                )
        for channel in self.socket.metadata['channels']:
            if channel['is_archived']:
                continue
            if not self.get_channel(channel['id']):
                horse.models.Channel(
                    string_id=channel['id'],
                    name=channel['name']
                )

    def load_bridles(self):
        self.bridles = {
            "event": [],
            "command": [],
            "listener": [],
            "webhook": {
                "GET": [],
                "POST": []
            }
        }
        self.bridle_models = []

        logging.info("")
        logging.info('<<<< START Loading Bridles >>>>')
        for bridle_path in horse.config.BRIDLES:
            logging.info("")
            logging.info('Loading Bridle: {0}'.format(bridle_path))
            if isinstance(bridle_path, basestring):
                bridle_module = importlib.import_module(
                    bridle_path.replace('horse.ext.', 'horse_')
                )
            elif isinstance(bridle_path, types.ModuleType):
                bridle_module = bridle_path
            else:
                logging.warning("\tFailed: {0}".format(bridle_path))
                continue

            reserved_commands = []

            for cls in bridle_module.__dict__.values():

                if not inspect.isclass(cls):
                    continue

                if issubclass(cls, horse.models.Model):

                    if cls.Meta.abstract:
                        continue

                    logging.info('\tModel: {0}'.format(cls.__name__))
                    self.bridle_models.append(cls)

                elif issubclass(cls, horse.bridles.base.Bridle):

                    if cls.Meta.abstract:
                        continue

                    if issubclass(cls, horse.bridles.base.CommandBridle):

                        if cls.Meta.command in reserved_commands:
                            logging.warning(
                                '\tIgnored: {0} -> {1}'.format(
                                    cls.Meta.command,
                                    cls.__name__
                                )
                            )
                            continue

                        logging.info('\tCommand: {0} -> {1}'.format(
                            cls.Meta.command,
                            cls.__name__
                        ))
                        reserved_commands.append(cls.Meta.command)
                        self.bridles['command'].append(cls(self))

                    elif issubclass(cls, horse.bridles.base.ListenerBridle):
                        logging.info('\tListener: {0} -> {1}'.format(
                            cls.Meta.regex,
                            cls.__name__
                        ))
                        self.bridles['listener'].append(cls(self))

                    elif issubclass(cls, horse.bridles.base.WebhookBridle):
                        logging.info('\tWebhook: {0}:{1} -> {2}'.format(
                            cls.Meta.method,
                            cls.Meta.path,
                            cls.__name__
                        ))
                        self.bridles['webhook'][cls.Meta.method].append(
                            cls(self)
                        )

                    elif issubclass(cls, horse.bridles.base.EventBridle):
                        logging.info('\tEvent: {0} -> {1}'.format(
                            cls.Meta.event,
                            cls.__name__
                        ))
                        self.bridles['event'].append(cls(self))

        logging.info("")
        horse.database.verify_models(self.bridle_models)
        logging.info('<<<< END Loading Bridles >>>>')
        logging.info("")

    def handle_socket(self, data):
        logging.info('Handling socket event: Type={0}'.format(data['type']))

        if 'user' in data:
            if isinstance(data['user'], basestring):
                user = self.get_user(data['user'])
            elif isinstance(data['user'], dict):
                user = self.get_user(data['user']['id'])
                if user is None:
                    user = data['user']
                    user = horse.models.User(
                        string_id=user['id'],
                        username=user['name'],
                        real_name=user['profile']['real_name_normalized'],
                        image=user['profile']['image_72']
                    )
            if user is not None:
                data['user'] = user

        if 'channel' in data:
            channel = self.get_channel(data['channel'])
            if channel is not None:
                data['channel'] = channel

        if data['type'] == 'message' and 'subtype' not in data:
            self.handle_message(data)

        else:
            for bridle in self.bridles['event']:
                if data['type'] == bridle.Meta.event:
                    bridle.execute(data)

    def handle_message(self, data):
        logging.debug(data)
        for bridle in self.bridles['listener']:
            match = bridle._pattern.search(data['text'])
            if match:
                bridle.execute(
                    data['user'],
                    data['channel'],
                    match,
                    data['text']
                )

    def handle_http_get(self, path, params):
        logging.info('Handling GET webhook: {0}'.format(path))
        for bridle in self.bridles['webhook']['GET']:
            if bridle.Meta.path == path:
                return bridle.execute(path, params)

        return ""

    def handle_http_post(self, path, data):
        logging.info('Handling POST webhook: {0}'.format(path))
        for bridle in self.bridles['webhook']['POST']:
            if bridle.Meta.path == path:
                return bridle.execute(path, data)

        return ""

    def handle_command(self, data):
        logging.info('Handling command event')

        if (
            'text' not in data or
            'user_id' not in data or
            'channel_id' not in data
        ):
            logging.error('Malformed command request')
            return 'Malformed command request'

        if 'user_id' in data:
            user = self.get_user(data['user_id'])

        if 'channel_id' in data:
            channel = self.get_channel(data['channel_id'])

        operands = data['text'].split(' ')
        command = operands[0]
        for bridle in self.bridles['command']:

            if bridle.Meta.command == command:
                try:
                    return bridle.execute(
                        user,
                        channel,
                        operands[1:]
                    )
                except:
                    traceback.print_exc()
                    return "Failure: '{0}'".format(data['text'])

        return "Unknown command!"



import logging

HOSTNAME = ''
DATABASE_URI = ''

SLACK_API_TOKEN = ''
FORECASTIO_API_TOKEN = ''
IMGUR_API_TOKEN = ''


BRIDLES = [
    'horse.bridles.core.help',
    'horse.bridles.core.uptime',
    'horse.bridles.core.weather',
]

logging.basicConfig(
    format='[%(asctime)s]: {%(threadName)s} (%(levelname)s) %(message)s',
    level=logging.DEBUG
)



"""A setuptools based setup module for horse.

https://github.com/grieve/horse
"""

from os import path
from setuptools import setup
from setuptools import find_packages

here = path.abspath(path.dirname(__file__))

setup(
    name='horse',
    version='0.2.16',

    description='A chat bot framework for Slack',
    long_description="""
""",

    url='https://github.com/grieve/horse',

    author='Ryan Grieve',
    author_email='me@ryangrieve.com',

    license='MIT',

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Communications :: Chat',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
    ],

    keywords='chatbot slack chat bot communication',

    packages=find_packages(exclude=['contrib', 'docs', 'tests']),

    install_requires=[
        'six',
        'websocket-client',
        'requests',
        'sqlobject',
        'flask',
    ],

    extras_require={
        'dev': [],
        'test': [],
    },

    package_data={},
    data_files=[],
    entry_points={},
)



from cStringIO import StringIO
from base64 import b64encode

import requests

from .. import config


def post_from_path(image, caption=None):
    image_file = open(image, 'r')
    return post_from_file(image_file)


def post_from_file(image_file, caption=None):
    data = b64encode(image_file.read())
    return post_b64_data(data, caption)


def post_from_pil(image, caption=None):
    output = StringIO()
    image.save(output, format="PNG")
    output.seek(0)
    return post_from_file(output, caption)


def post_string_data(data, caption=None):
    b64_data = b64encode(data)
    return post_b64_data(b64_data, caption)


def post_b64_data(b64_data, caption=None):
    if caption is None:
        caption = "Autogenerated by a horse..."

    payload = {
        'album_id': None,
        'image': b64_data,
        'title': caption,
        'description': None
    }

    resp = requests.post(
        "https://api.imgur.com/3/image",
        data=payload,
        headers={
            "Authorization": "Client-ID {0}".format(config.IMGUR_API_TOKEN)
        }
    )
    return resp.json()['data']['link']



import six
import sqlobject


class ModelBase(sqlobject.SQLObject.__metaclass__):

    class DefaultMeta(object):
        pass

    def __new__(cls, name, bases, attrs):
        meta = attrs.pop('Meta', None)
        if len(bases) == 0:
            bases = (sqlobject.SQLObject,)
        new_class = super(ModelBase, cls).__new__(cls, name, bases, attrs)
        if meta is None:
            meta = ModelBase.DefaultMeta
        if 'abstract' not in meta.__dict__:
            meta.abstract = False
        new_class.Meta = meta
        return new_class


class Model(six.with_metaclass(ModelBase)):
    created = sqlobject.DateTimeCol(default=sqlobject.DateTimeCol.now())


class User(Model):
    string_id = sqlobject.StringCol()
    username = sqlobject.StringCol()
    real_name = sqlobject.StringCol()
    image = sqlobject.StringCol()

    def __str__(self):
        return self.string_id


class Channel(Model):
    string_id = sqlobject.StringCol()
    name = sqlobject.StringCol()

    def __str__(self):
        return self.string_id
